#include <gccore.h>
#include <malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ogcsys.h>
#include <unistd.h>
#include <time.h>
#include <ogc/mic.h>

//#define DEBUG_MIC
#ifdef DEBUG_MIC
#	include <debug.h>
#endif

#define OUTPUT_AUDIO
#ifdef OUTPUT_AUDIO
#	include <aesndlib.h>
#endif

#include "ConsoleHelper.h"

static void *xfb = NULL;

u32 first_frame = 1;
GXRModeObj *rmode;
void (*reload)() = (void(*)())0x80001800;

void InitCrap()
{
	VIDEO_Init();
	rmode = VIDEO_GetPreferredMode(NULL);

	PAD_Init();
	
	xfb = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
		
	VIDEO_Configure(rmode);
		
	VIDEO_SetNextFramebuffer(xfb);
	VIDEO_SetBlack(FALSE);
	VIDEO_Flush();
	VIDEO_WaitVSync();
	if(rmode->viTVMode&VI_NON_INTERLACE)
		VIDEO_WaitVSync();
	console_init(xfb,20,64,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*2);
	VIDEO_SetNextFramebuffer(xfb);
}

enum
{
    MIC_PROBE = 0,
    MIC_MOUNT,
    MIC_SETPARAMS,
    MIC_START,
    MIC_WAIT,
    MIC_AVAILABLE
};
    
typedef struct
{
    s32 status;
    s32 currStage;
    s32 nextStage;
    s16 *buffer;
#ifdef OUTPUT_AUDIO
    AESNDPB *voice;
	s32 currBottom;
#endif
} MICData;

static MICData  __mic_data[2] =
{
    {MIC_RESULT_NOCARD, MIC_PROBE, MIC_PROBE, NULL
	#ifdef OUTPUT_AUDIO
	, NULL, 0
	#endif
	},
    {MIC_RESULT_NOCARD, MIC_PROBE, MIC_PROBE, NULL
	#ifdef OUTPUT_AUDIO
	, NULL, 0
	#endif
	},
};

static void __MICDetachCB(s32 chan, s32 result)
{
    __mic_data[chan].status = MIC_RESULT_NOCARD;
}

static void MICDeinit(s32 chan)
{
	MICData *md = &__mic_data[chan];
	
	MICUnmount(chan);
#ifdef OUTPUT_AUDIO
	if (md->voice)
	{
		AESND_SetVoiceStop(md->voice, true);
		AESND_FreeVoice(md->voice);
		md->voice = NULL;
	}
#endif
	
	if (md->buffer)
	{
		free(md->buffer);
		md->buffer = NULL;
	}
	
	md->currStage = MIC_PROBE;
}

#ifdef OUTPUT_AUDIO
// hw ring buffer -> 12KB driver ringbuffer -> nicely sized copy for aesnd
#define AESNDBUFSIZE	DSP_STREAMBUFFER_SIZE * 2 * sizeof(s16)
#define AESNDNUMBUFS	2
static s16 aesnd_buffers[AESNDNUMBUFS][AESNDBUFSIZE / sizeof(s16)];
static int aesnd_buff_fill = 0;
static int aesnd_buff_play = 0;
static u32 aesnd_buff_off = 0;
static bool aesnd_ready = false;

// called each time the hw buffer is read
static void fill_callback(s32 chan, s32 result)
{
	s32 num_bytes, num_samples;
	s16 samples[128 / sizeof(s16)];
	s32 *bottom = &__mic_data[chan].currBottom;
	
	MICGetBuffsize(chan, &num_bytes);
	num_samples = num_bytes / sizeof(s16);
	MICGetSamples(chan, samples, *bottom, num_samples);
	memcpy(&aesnd_buffers[aesnd_buff_fill][aesnd_buff_off / sizeof(s16)], samples, num_bytes);
	*bottom = MICUpdateIndex(chan, *bottom, num_samples);
	aesnd_buff_off += num_bytes;
	
	if (aesnd_buff_off >= AESNDBUFSIZE)
	{
		aesnd_buff_off = 0;
		aesnd_ready = true;
		aesnd_buff_play = aesnd_buff_fill;
		aesnd_buff_fill = (aesnd_buff_fill + 1) % AESNDNUMBUFS;
		memset(aesnd_buffers[aesnd_buff_fill], 0, AESNDBUFSIZE);
	}
}

static void aesnd_callback(AESNDPB *pb, u32 state)
{
	if (state == VOICE_STATE_STREAM && aesnd_ready)
	{
		aesnd_ready = false;
		
		AESND_SetVoiceBuffer(pb, aesnd_buffers[aesnd_buff_play], AESNDBUFSIZE);
	}
}
#endif

static void __MICUpdate(void)
{
    static u32 old_button[2] = {0, 0};
    
    MICData*  md;
    s32       chan;
    s32       result;
    u32       button;
    
    for (chan = 0; chan < 2; chan++)
    {
        md = &__mic_data[chan];
        
        switch(md->currStage)
        {
		case MIC_PROBE:
			//
			// Check whether microphone device is attached 
			//
			result = MICProbeEx(chan);
			if (result == MIC_RESULT_READY)
			{
				//OSReport("chan %ld : find mic\n", chan);
				md->currStage = MIC_MOUNT;
			}
			break;
			
		case MIC_MOUNT:
			//
			// Try mounting microphone device 
			//
			md->buffer = (s16*)malloc(MIC_RINGBUFF_SIZE);
			//OSReport("chan %ld : try to mount mic\n", chan);
			result = MICMountAsync(chan,
								   md->buffer,
								   MIC_RINGBUFF_SIZE,
								   __MICDetachCB,
								   NULL);
			switch(result)
			{
				case MIC_RESULT_READY:
					md->currStage = MIC_WAIT;
					md->nextStage = MIC_SETPARAMS;
					break;
				case MIC_RESULT_BUSY:
					break;
				default:
					goto ERROR;
					break;
			}
			break;
					
		case MIC_SETPARAMS:
			//
			// Try setting the microphone device parameters
			//
			//OSReport("chan %ld : try to set params\n", chan);
			result = MICSetParamsAsync(chan,
									   32,
									   44100,
									   15,
									   NULL);
			switch(result)
			{
				case MIC_RESULT_READY:
					md->currStage = MIC_WAIT;
					md->nextStage = MIC_START;
					break;
				case MIC_RESULT_BUSY:
					break;
				default:
					goto ERROR;
					break;
			}
			break;
					
		case MIC_START:
			//
			// Try starting the microphone device
			//
			#ifdef OUTPUT_AUDIO
			MICSetTxCallback(chan, fill_callback);
			#endif

			result = MICStartAsync(chan, NULL);
			//OSReport("chan %ld : try to start mic\n", chan);
			
			#ifdef OUTPUT_AUDIO
			s32 rate;
			MICGetRate(chan, &rate);
			
			// setup pb
			if (!md->voice)
			{
				md->voice = AESND_AllocateVoice(aesnd_callback);
				if (md->voice)
				{
					AESND_SetVoiceFormat(md->voice, VOICE_MONO16);
					AESND_SetVoiceFrequency(md->voice, rate);
					AESND_SetVoiceVolume(md->voice, 0xFF, 0xFF);
					AESND_SetVoiceStream(md->voice, true);
				}
				else
					result = MIC_RESULT_INVALID_STATE;
			}
			else
				result = MIC_RESULT_INVALID_STATE;
			#endif
			
			switch(result)
			{
				case MIC_RESULT_READY:
					md->currStage = MIC_WAIT;
					md->nextStage = MIC_AVAILABLE;
					#ifdef OUTPUT_AUDIO
					// start it
					AESND_SetVoiceStop(md->voice, false);
					#endif
					break;
				case MIC_RESULT_BUSY:
					break;
				default:
					goto ERROR;
					break;
			}

		case MIC_WAIT:
			//
			// Wait for end of the various microphone device tests
			//
			result = MICGetResultCode(chan);
			switch(result)
			{
				case MIC_RESULT_READY:
					//OSReport("chan %ld : done\n", chan);
					md->status    = MIC_RESULT_READY;
					md->currStage = md->nextStage;
					break;
				case MIC_RESULT_BUSY:
					break;
				default:
					goto ERROR;
					break;
			}
			break;
					
		case MIC_AVAILABLE:

			//
			// Get microphone button state -> switch effect
			//
			MICGetButton(chan, &button);
			button &= MIC_BUTTON_TALK;
			if((old_button[chan] ^ button) & button)
			{
			#ifdef DEBUG_MIC
				_break();
			#endif
			}
			old_button[chan] = button;
			
			//                
			// Microphone device is operating
			//
			if (md->status != MIC_RESULT_READY)
			{
				//
				// Microphone device is detached
				//
				//OSReport("chan %ld : detached mic\n", chan);
				MICDeinit(chan);
			}
			break;
        }

        continue;

ERROR:
        if (result == MIC_RESULT_WRONGDEVICE ||
			result == MIC_RESULT_NOCARD)
        {
			MICDeinit(chan);
        }
    }
}

static const char *GetResultString(s32 result)
{
	switch (result)
	{
	case MIC_RESULT_UNLOCKED: return "UNLOCKED";
	case MIC_RESULT_READY: return "READY";
	case MIC_RESULT_BUSY: return "BUSY";
	case MIC_RESULT_WRONGDEVICE: return "WRONGDEVICE";
	case MIC_RESULT_NOCARD: return "NOCARD";
	case MIC_RESULT_INVALID_STATE: return "INVALID_STATE";
	case MIC_RESULT_FATAL_ERROR: return "FATAL_ERROR";
	default: return "UNKNOWN";
	}
}

// wii ~22030 ticks
// gc  ~14700 ticks
//extern u32 micdifftime;

static void PrintInfos()
{
	int columns = 0, rows = 0;
	CON_GetMetrics(&columns, &rows);
	
	const int y_off = 5;
	const int buff_y_off = y_off * 2 + 1;
	const int buff_rows = (rows - buff_y_off) / 2;
	
	for (int chan = 0; chan < 2; chan++)
	{
		int y = 0;
		
		u32 pattern, button, id, status;
		s32 ringsize, buffsize, rate, gain;
		
		MICGetOut(chan, &pattern);
		MICGetParams(chan, &buffsize, &rate, &gain);
		MICGetButton(chan, &button);
		MICGetDeviceID(chan, &id);
		MICGetStatus(chan, &status);
		MICGetRingbuffsize(chan, &ringsize);
		
		VIDEO_WaitVSync();
		
		++y;
		CON_PrintRow(2, y + y_off * chan,
			"MIC %i probe: %s status: %s %s %s",
			chan, GetResultString(MICProbeEx(chan)), GetResultString(MICGetResultCode(chan)),
			MICIsAttached(chan) ? "attached" : "", MICIsActive(chan) ? "active" : "");
		++y;
		CON_PrintRow(2, y + y_off * chan,
			"samples left: %4i errors: %i",
			MICGetSamplesLeft(chan, 0), MICGetErrorCount(chan));
		++y;
		CON_PrintRow(2, y + y_off * chan,
			"out: %x ringsize: %i buffsize: %i rate: %i gain: %i",
			pattern, ringsize, buffsize, rate, gain);
		++y;
		CON_PrintRow(2, y + y_off * chan,
			"id: %x button: %02x status: %04x", id, button, status);
		++y;
		CON_PrintRow(2, y + y_off * chan, "------------------------------------");
		
		MICData *md = &__mic_data[chan];
		
		if (md->buffer == NULL)
		{
			// blank bad data
			for (int r = y_off * chan + 2; r < y_off * chan + 4; r++)
			{
				CON_BlankRow(r);
			}
			for (int i = 0; i < buff_rows - chan; i++)
			{
				const int r = i + buff_y_off + buff_rows * chan + chan;
				CON_BlankRow(r);
			}
			continue;
		}
			
		for (int i = 0; i < buff_rows - chan; i++)
		{
			int c = 0;
			const int r = i + buff_y_off + buff_rows * chan + chan;
			
			for (int bi = 0; bi < 16; bi++)
			{
				CON_Printf(c, r, "%04x ", (u16)md->buffer[i * 16 + bi]);
				c += 5;
			}
		}
		
		if (chan == 0)
			CON_PrintRow(2, buff_rows + buff_y_off, "------------------------------------");
		
	}
}

int main()
{
	InitCrap();
	
	MICInit();
	
#ifdef OUTPUT_AUDIO
	AESND_Init();
#endif

#ifdef DEBUG_MIC
	DEBUG_Init(GDBSTUB_DEVICE_USB, 1);
	_break();
#endif

	CON_Clear();

	while (1)
	{
		PAD_ScanPads();
		if (PAD_ButtonsDown(0) & PAD_BUTTON_START)
			reload();

		__MICUpdate();
		
		PrintInfos();
	}
}
